package com.sillycycle.bagleyd.life2d;

/* Code for selection spliced in is due to Alex Shvets */
/* http://home.comcast.net/~shvets/Programs/Examples/GameLife/life2.html */
// Life2DMatrix.java

import java.awt.Rectangle;
import java.io.Externalizable;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.sillycycle.bagleyd.util.CellInfo;

/* Pointer design (do not need buffer copy):
Start with live list,
  set grid array and linked list
  assume toggles on grid array all 0  (setCell)
for each live cell in linked list
  create a dead list if toggle 0 and grid array  off
  set toggle on grid
for each dead cell if comes alive keep, else remove
for each live cell set toggle if dies
for each live cell update according to toggle
  if dies erase cell remove.
  if lives update age, draw if necessary.
for each dead cell update to live cell.
*/

public class Life2DMatrix implements Externalizable {
	private static final long serialVersionUID = 42L;

	private int columns, rows;
	private int area;

	private byte[] cells;
	public transient List<CellInfo> list, deadList;

	// must exist for deserialization process
	public Life2DMatrix() {
	}

	public Life2DMatrix(int columns, int rows) {
		this.columns = columns;
		this.rows = rows;
		area = columns * rows;
		list = new LinkedList<>();
		deadList = new LinkedList<>();
		init();
	}

	public final void init() {
		cells = new byte[area];
		clear();
	}

	public void clear() {
		for (int position = 0; position < area; position++)
			cells[position] = (byte) 0; // Empty
	}

	public void clearLiveCells() {
		for (Iterator<CellInfo> iterator = list.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int position = cellInfo.getPosition();
			cells[position] = 0;
			iterator.remove();
		}
	}

	// O(n)
	public void clearCell(int position) {
		for (Iterator<CellInfo> iterator = list.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int p = cellInfo.getPosition();
			if (position == p)
				iterator.remove();
			//break; // commented to clear duplications
		}
		cells[position] = 0;
	}

	public int getColumns() {
		return columns;
	}

	public int getRows() {
		return rows;
	}

	public byte getCell(int position) {
		byte value = cells[position];

		if (value == -128)
			return 0;
		else if (value < 0)
			return (byte) -value;
		else
			return value;
	}

	public boolean getCellBit(int position) {
		return (cells[position] < 0);
	}

	public void setArrayCell(int position, byte cell) {
		cells[position] = cell;
	}

	public void setCell(int position, byte cell) {
		if (cells[position] == 0) {
			list.add(new CellInfo(position, cell));
		} else if (cell == 0) {
			for (Iterator<CellInfo> iterator = list.iterator();
					iterator.hasNext();) {
				CellInfo cellInfo = iterator.next();
				int p = cellInfo.getPosition();
				if (position == p)
					iterator.remove();
				//break; // commented to clear duplications
			}
		} else {
			for (CellInfo cellInfo : list) {
				int p = cellInfo.getPosition();
				if (position == p) {
					cellInfo.setCell(cell);
				}
			}
		}
		cells[position] = cell;
	}

	public void setDeadCell(int position) {
		if (cells[position] == 0) {
			setCellBit(position);
			CellInfo cellInfo = new CellInfo(position);
			deadList.add(cellInfo);
//System.out.println(" " + position + " " + getCellBit(position));
		}
	}

	public void setCellBit(int position) {
		byte value = cells[position];

		if (value == 0)
			value = -128;
		else if (value > 0)
			value = (byte) -value;
		cells[position] = value;
	}

	public void resetCellBit(int position) {
		byte value = cells[position];

		if (value == -128)
			value = 0;
		else if (value < 0)
			value = (byte) -value;
		cells[position] = value;
	}

	public byte getCell(int column, int row) {
		if ((row >= 0) && (row < rows) &&
				(column >= 0) && (column < columns))
			return getCell(row * columns + column);
		return 0;
	}

	// To be used so do not add to dead list for cells already added
	public boolean getCellBit(int column, int row) {
		if ((row >= 0) && (row < rows) &&
				(column >= 0) && (column < columns))
			return getCellBit(row * columns + column);
		return true;
	}

	public void setCell(int column, int row, byte cell) {
		if ((row >= 0) && (row < rows) &&
				(column >= 0) && (column < columns)) {
			int position = row * columns + column;
			if (cell == 0) {
				clearCell(position);
			} else {
				setCell(position, cell);
			}
		}
	}

	// multiple loops are ok for SubMatrix stuff, time not an issue

	public Life2DMatrix getSubMatrix(Rectangle r) {
		Life2DMatrix selection = new Life2DMatrix(r.width, r.height);

		for (int j = 0; j < r.height; j++)
		for (int i = 0; i < r.width; i++) {
			selection.setCell(i, j,
				getCell(i + r.x, j + r.y));
		}
		return selection;
	}

	public void setSubMatrix(Life2DMatrix selection, Rectangle r) {
		for (int j = 0; j < r.height; j++)
		for (int i = 0; i < r.width; i++) {
			try {
				setCell(i + r.x, j + r.y,
					selection.getCell(i, j));
			} catch (ArrayIndexOutOfBoundsException e) {
				// ignore
			}
		}
	}

	public void clearSubMatrix(Rectangle r) {
		for (int j = r.y; j < r.y + r.height; j++)
		for (int i = r.x; i < r.x + r.width; i++) {
			clearCell(j * columns + i);
		}
	}

	private static int ceil2(int z) {
		if (z >= 0)
			return (z + 1) / 2;
		return z / 2;
	}

	/* The center point of rectangle (more or less) is used
	 * to rotate about.
	 */
	void rotate(Life2DMatrix selection, Rectangle r,
			int polygon, boolean cw) {
		for (int j = 0; j < r.height; j++)
		for (int i = 0; i < r.width; i++) {
			switch (polygon) {
			case 3:
				if (cw)
					setCell(2 * (2 * r.height / 3) + ceil2(i - j) - j - 2 + r.x,
						(i + j) / 2 + 1 + r.y,
						selection.getCell(i, j));
				else
					setCell(j + (i + j) / 2 - 3 + r.x,
						2 * (r.width / 4) - ceil2(i - j) + r.y,
						selection.getCell(i, j));
				break;
			case 4:
				if (cw)
					setCell(r.height - j - 1 + r.x,
						i + r.y,
						selection.getCell(i, j));
				else
					setCell(j + r.x,
						r.width - i - 1 + r.y,
						selection.getCell(i, j));
				break;
			case 5:
				int mod = (i + 2 * j) % 4;
				int tcol = j * 2;
				int trow = i / 2;

				if (mod == 1 || mod == 2)
					tcol++;
				if (cw)
					setCell(4 * (r.height / 2) - tcol - 1 + r.x,
						trow + r.y,
						selection.getCell(i, j));
				else
					setCell(tcol + r.x,
						2 * (r.width / 4) - trow - 1 + r.y,
						selection.getCell(i, j));
				break;
			case 6:
				if (cw)
					setCell(2 * (r.height / 2) - j + (i + 1 + j / 2) / 2 + r.x,
						i + j / 2 + r.y,
						selection.getCell(i, j));
				else
					setCell(j + (i + 1 - ((j + 1) / 2)) / 2 + r.x,
						2 * (r.width / 2) - i + (j + 1) / 2 - 2 + r.y,
						selection.getCell(i, j));
				break;
			default:
			}
		}
	}

	/* Wall means you reflecting in a wall mirror,
	 * else it means you are reflecting off ground.
	 */
	void reflect(Life2DMatrix selection, Rectangle r,
			int polygon, boolean wall) {

		for (int j = 0; j < r.height; j++)
		for (int i = 0; i < r.width; i++) {
			switch (polygon) {
			case 3:
				if (wall) {
					int evenX = 1 - (r.width & 1);

					setCell(r.width - i - 1 + evenX + r.x,
						j + r.y,
						selection.getCell(i, j));
				} else {
					int oddY = (r.height & 1);

					setCell(i + r.x,
						r.height - j - 1 + oddY + r.y,
						selection.getCell(i, j));
				}
				break;
			case 4:
				if (wall)
					setCell(r.width - i - 1 + r.x,
						j + r.y,
						selection.getCell(i, j));
				else
					setCell(i + r.x,
						r.height - j - 1 + r.y,
						selection.getCell(i, j));
				break;
			case 5:
				int mod = (i + 2 * j) % 4;

				if (wall) {
					int oddX = ((r.width / 2) & 1);
					int tcol = r.width - i;

					if (mod == 1)
						tcol += 2;
					else if (mod == 2 || mod == 3)
						tcol++;
					setCell(tcol + r.x + 2 * oddX,
						j + r.y,
						selection.getCell(i, j));
				} else {
					int oddY = (r.height & 1);
					int tcol = i;
					int trow = r.height - j + oddY;

					if (mod == 0)
						tcol++;
					else if (mod == 1)
						tcol--;
					setCell(tcol + r.x,
						trow + r.y,
						selection.getCell(i, j));
				}
				break;
			case 6:
				if (wall)
					setCell(r.width - i - 1 - (((j & 1) == 1) ? 0 : 1) + r.x,
						j + r.y,
						selection.getCell(i, j));
				else
					setCell(i + r.x,
						r.height - j + r.y - 2,
						selection.getCell(i, j));
				break;
			default:
			}
		}
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();

		for (int j = 0; j < rows; j++) {
			for (int i = 0; i < columns; i++)
				sb.append(cells[j * columns + i]);
			sb.append("\n");
		}
		sb.append("\n");
		return sb.toString();
	}


	// implements Externalizable

	@Override
	public void writeExternal(ObjectOutput out) throws IOException {
		out.writeByte(columns);
		out.writeByte(rows);
		for (int j = 0; j < rows; j++)
		for (int i = 0; i < columns; i++) {
			if (cells[j * columns + i] == 1) {
				out.writeByte(i);
				out.writeByte(j);
			}
		}
	}

	@Override
	public void readExternal(ObjectInput in)
			throws IOException, ClassNotFoundException {
		columns = in.readByte();
		rows = in.readByte();
		init();
		while (true) {
			try {
				int i = in.readByte();
				int j = in.readByte();
				cells[j * columns + i] = 1;
			} catch (IOException e)	{
				break;
			}
		}
	}

	public static void main(String[] args) {
		System.out.println("Constructing object:");
		Life2DMatrix m = new Life2DMatrix(9, 8);
		m.setCell(2, 3, (byte) 1);
		m.setCell(3, 3, (byte) 1);
		m.setCell(4, 3, (byte) 1);
		m.setCell(4, 4, (byte) 1);
		m.setCell(3, 5, (byte) 1);
		Life2DMatrix n;

		System.out.println("Saving object:");
		try (ObjectOutputStream out = new ObjectOutputStream(
				new FileOutputStream("junk.out"))) {
			out.writeObject(m);
		} catch (Exception e) {
			e.printStackTrace();
		}
		System.out.println("Recovering object:");
		// Now get it back
		try (ObjectInputStream in = new ObjectInputStream(
				new FileInputStream("junk.out"))) {
			n = (Life2DMatrix)in.readObject();
			System.out.println(n.toString());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
